from pathlib import Path
import pytest

from espnet_model_zoo.downloader import cmd_download
from espnet_model_zoo.downloader import cmd_query
from espnet_model_zoo.downloader import download
from espnet_model_zoo.downloader import ModelDownloader


def test_download():
    download("http://example.com", "index.html")


def test_update_model_table(tmp_path):
    d = ModelDownloader(tmp_path)
    d.update_model_table()


def test_get_data_frame():
    d = ModelDownloader()
    d.get_data_frame()


def test_new_cachedir(tmp_path):
    ModelDownloader(tmp_path)


def test_download_and_unpack_names_with_condition():
    d = ModelDownloader()
    d.query("name", task="asr")


def test_get_model_names_and_urls():
    d = ModelDownloader()
    d.query(["name", "url"], task="asr")


def test_get_model_names_non_matching():
    d = ModelDownloader()
    assert d.query("name", task="dummy") == []


def test_download_and_unpack_with_url():
    d = ModelDownloader()
    d.download_and_unpack("https://zenodo.org/record/3951842/files/test.zip?download=1")


def test_download_and_unpack_with_name():
    d = ModelDownloader()
    d.download_and_unpack("test")


def test_download_and_unpack_no_inputting():
    d = ModelDownloader()
    with pytest.raises(TypeError):
        d.download_and_unpack()


def test_download_and_unpack_non_matching():
    d = ModelDownloader()
    with pytest.raises(RuntimeError):
        d.download_and_unpack(task="dummy")


def test_download_and_unpack_local_file():
    d = ModelDownloader()
    path = d.download("test")
    d.download_and_unpack(path)


def test_download_and_clean_cache():
    d = ModelDownloader()
    d.download_and_unpack("test")
    p = d.download("test")
    d.clean_cache("test")
    assert not Path(p).exists()


def test_cmd_download():
    cmd_download(["test"])


def test_query():
    cmd_query([])



import argparse
from distutils.util import strtobool
import hashlib
import os
from pathlib import Path
import re
import shutil
import tempfile
from typing import Dict
from typing import List
from typing import Sequence
from typing import Tuple
from typing import Union
import warnings

from filelock import FileLock
from huggingface_hub import snapshot_download
import pandas as pd
import requests
from tqdm import tqdm
import yaml

from espnet2.main_funcs.pack_funcs import find_path_and_change_it_recursive
from espnet2.main_funcs.pack_funcs import get_dict_from_cache
from espnet2.main_funcs.pack_funcs import unpack


MODELS_URL = (
    "https://raw.githubusercontent.com/espnet/espnet_model_zoo/master/"
    "espnet_model_zoo/table.csv"
)


URL_REGEX = re.compile(
    r"^(?:http|ftp)s?://"  # http:// or https://
    r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)"
    r"+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"  # domain...
    r"localhost|"  # localhost...
    r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # ...or ip
    r"(?::\d+)?"  # optional port
    r"(?:/?|[/?]\S+)$",
    re.IGNORECASE,
)


def is_url(url: str) -> bool:
    return re.match(URL_REGEX, url) is not None


def str_to_hash(string: Union[str, Path]) -> str:
    return hashlib.md5(str(string).encode("utf-8")).hexdigest()


def download(
    url, output_path, retry: int = 3, chunk_size: int = 8192, quiet: bool = False
):
    # Set retry
    session = requests.Session()
    session.mount("http://", requests.adapters.HTTPAdapter(max_retries=retry))
    session.mount("https://", requests.adapters.HTTPAdapter(max_retries=retry))

    # Timeout
    response = session.get(url=url, stream=True, timeout=(10.0, 30.0))
    file_size = int(response.headers["content-length"])

    # Raise error when connection error
    response.raise_for_status()

    # Write in temporary file
    with tempfile.TemporaryDirectory() as d:
        with (Path(d) / "tmp").open("wb") as f:
            if quiet:
                for chunk in response.iter_content(chunk_size=chunk_size):
                    if chunk:
                        f.write(chunk)
            else:
                with tqdm(
                    desc=url,
                    total=file_size,
                    unit="B",
                    unit_scale=True,
                    unit_divisor=1024,
                ) as pbar:
                    for chunk in response.iter_content(chunk_size=chunk_size):
                        if chunk:
                            f.write(chunk)
                            pbar.update(len(chunk))

        Path(output_path).parent.mkdir(parents=True, exist_ok=True)
        shutil.move(Path(d) / "tmp", output_path)


class ModelDownloader:
    """Download model from zenodo and unpack."""

    def __init__(self, cachedir: Union[Path, str] = None):
        if cachedir is None:
            # The default path is the directory of this module
            cachedir = Path(__file__).parent
            # If not having write permission, fallback to homedir
            if not os.access(cachedir, os.W_OK):
                cachedir = Path.home() / ".cache" / "espnet_model_zoo"
        else:
            cachedir = Path(cachedir).expanduser().absolute()
        cachedir.mkdir(parents=True, exist_ok=True)

        csv = Path(__file__).parent / "table.csv"
        if not csv.exists():
            download(MODELS_URL, csv)

        self.cachedir = cachedir
        self.csv = csv
        self.data_frame = pd.read_csv(csv, dtype=str)

    def get_data_frame(self):
        return self.data_frame

    def update_model_table(self):
        download(MODELS_URL, self.csv)

    def clean_cache(self, name: str = None, version: int = -1, **kwargs: str):
        url = self.get_url(name=name, version=version, **kwargs)
        outdir = self.cachedir / str_to_hash(url)
        shutil.rmtree(outdir)

    def query(
        self, key: Union[Sequence[str]] = "name", **kwargs
    ) -> List[Union[str, Tuple[str]]]:
        conditions = None
        for k, v in kwargs.items():
            if k not in self.data_frame:
                warnings.warn(
                    f"Invalid key: {k}: Available keys:\n"
                    f"{list(self.data_frame.keys())}"
                )
                continue
            condition = self.data_frame[k] == v
            if conditions is None:
                conditions = condition
            else:
                conditions &= condition

        if conditions is not None:
            df = self.data_frame[conditions]
        else:
            df = self.data_frame

        if len(df) == 0:
            return []
        else:
            if isinstance(key, (tuple, list)):
                return list(zip(*[df[k] for k in key]))
            else:
                return list(df[key])

    def get_url(self, name: str = None, version: int = -1, **kwargs: str) -> str:
        if name is None and len(kwargs) == 0:
            raise TypeError("No arguments are given")

        if name is not None and is_url(name):
            # Specify the downloading link directly. "kwargs" are ignored in this case.
            url = name

        else:
            if name is not None:
                kwargs["name"] = name

            conditions = None
            for key, value in kwargs.items():
                condition = self.data_frame[key] == value
                if conditions is None:
                    conditions = condition
                else:
                    conditions &= condition

            if len(self.data_frame[conditions]) == 0:
                # If Specifying local file path
                if name is not None and Path(name).exists() and len(kwargs) == 1:
                    url = str(Path(name).absolute())

                else:
                    return "huggingface.co"
            else:
                urls = self.data_frame[conditions]["url"]
                if version < 0:
                    version = len(urls) + version
                url = list(urls)[version]
        return url

    @staticmethod
    def _get_file_name(url):
        ma = re.match(r"https://.*/([^/]*)\?download=[0-9]*$", url)
        if ma is not None:
            # URL e.g.
            # https://sandbox.zenodo.org/record/646767/files/asr_train_raw_bpe_valid.acc.best.zip?download=1
            a = ma.groups()[0]
            return a
        else:
            # If not Zenodo
            r = requests.head(url)
            if "Content-Disposition" in r.headers:
                # e.g. attachment; filename=asr_train_raw_bpe_valid.acc.best.zip
                for v in r.headers["Content-Disposition"].split(";"):
                    if "filename=" in v:
                        return v.split("filename=")[1].strip()

            # if not specified or some error happens
            return Path(url).name

    def unpack_local_file(self, name: str = None) -> Dict[str, Union[str, List[str]]]:
        if not Path(name).exists():
            raise FileNotFoundError(f"No such file or directory: {name}")

        warnings.warn(
            "Expanding a local model to the cachedir. "
            "If you'll move the file to another path, "
            "it's treated as a different model."
        )
        name = Path(name).absolute()

        outdir = self.cachedir / str_to_hash(name)
        filename = outdir / name.name
        outdir.mkdir(parents=True, exist_ok=True)

        if not filename.exists():
            if filename.is_symlink():
                filename.unlink()
            filename.symlink_to(name)

        # Skip unpacking if the cache exists
        meta_yaml = outdir / "meta.yaml"
        if meta_yaml.exists():
            info = get_dict_from_cache(meta_yaml)
            if info is not None:
                return info

        # Extract files from archived file
        return unpack(filename, outdir)

    def huggingface_download(
        self, name: str = None, version: int = -1, quiet: bool = False, **kwargs: str
    ) -> str:
        # Get huggingface_id from table.csv
        if name is None:
            names = self.query(key="name", **kwargs)
            if len(names) == 0:
                message = "Not found models:"
                for key, value in kwargs.items():
                    message += f" {key}={value}"
                raise RuntimeError(message)
            if version < 0:
                version = len(names) + version
            name = list(names)[version]

        if "@" in name:
            huggingface_id, revision = name.split("@", 1)
        else:
            huggingface_id = name
            revision = None

        return snapshot_download(
            huggingface_id,
            revision=revision,
            library_name="espnet",
            cache_dir=self.cachedir,
        )

    @staticmethod
    def _unpack_cache_dir_for_huggingfase(cache_dir: str):
        meta_yaml = Path(cache_dir) / "meta.yaml"
        lock_file = Path(cache_dir) / ".lock"
        flag_file = Path(cache_dir) / ".done"

        with meta_yaml.open("r", encoding="utf-8") as f:
            d = yaml.safe_load(f)
            assert isinstance(d, dict), type(d)

            yaml_files = d["yaml_files"]
            files = d["files"]
            assert isinstance(yaml_files, dict), type(yaml_files)
            assert isinstance(files, dict), type(files)

        # Rewrite yaml_files for first case
        with FileLock(lock_file):
            if not flag_file.exists():
                for key, value in yaml_files.items():
                    yaml_file = Path(cache_dir) / value
                    with yaml_file.open("r", encoding="utf-8") as f:
                        d = yaml.safe_load(f)
                        assert isinstance(d, dict), type(d)
                        for name in Path(cache_dir).glob("**/*"):
                            name = name.relative_to(Path(cache_dir))
                            d = find_path_and_change_it_recursive(
                                d, name, str(Path(cache_dir) / name)
                            )

                    with yaml_file.open("w", encoding="utf-8") as f:
                        yaml.safe_dump(d, f)

                with flag_file.open("w"):
                    pass

        retval = {}
        for key, value in list(yaml_files.items()) + list(files.items()):
            retval[key] = str(Path(cache_dir) / value)
        return retval

    def download(
        self, name: str = None, version: int = -1, quiet: bool = False, **kwargs: str
    ) -> str:
        url = self.get_url(name=name, version=version, **kwargs)

        # For huggingface compatibility
        if url in [
            "https://huggingface.co/",
            "https://huggingface.co",
            "huggingface.co",
        ]:
            # TODO(kamo): Support quiet
            cache_dir = self.huggingface_download(name=name, version=version, **kwargs)
            self._unpack_cache_dir_for_huggingfase(cache_dir)
            return cache_dir

        if not is_url(url) and Path(url).exists():
            return url

        outdir = self.cachedir / str_to_hash(url)
        filename = self._get_file_name(url)
        # Download the model file if not existing
        if not (outdir / filename).exists():
            download(url, outdir / filename, quiet=quiet)

            # Write the url for debugging
            with (outdir / "url").open("w", encoding="utf-8") as f:
                f.write(url)

            r = requests.head(url)
            if "Content-MD5" in r.headers:
                checksum = r.headers["Content-MD5"]

                # MD5 checksum
                sig = hashlib.md5()
                chunk_size = 8192
                with open(outdir / filename, "rb") as f:
                    while True:
                        chunk = f.read(chunk_size)
                        if len(chunk) == 0:
                            break
                        sig.update(chunk)

                if sig.hexdigest() != checksum:
                    Path(outdir / filename).unlink()
                    raise RuntimeError(f"Failed to download file: {url}")
            else:
                warnings.warn("Not validating checksum")
        return str(outdir / filename)

    def download_and_unpack(
        self, name: str = None, version: int = -1, quiet: bool = False, **kwargs: str
    ) -> Dict[str, Union[str, List[str]]]:
        url = self.get_url(name=name, version=version, **kwargs)
        if not is_url(url) and Path(url).exists():
            return self.unpack_local_file(url)

        # For huggingface compatibility
        if url in [
            "https://huggingface.co/",
            "https://huggingface.co",
            "huggingface.co",
        ]:
            # download_and_unpack and download are same if huggingface case
            # TODO(kamo): Support quiet
            cache_dir = self.huggingface_download(name=name, version=version, **kwargs)
            return self._unpack_cache_dir_for_huggingfase(cache_dir)

        # Unpack to <cachedir>/<hash> in order to give an unique name
        outdir = self.cachedir / str_to_hash(url)

        # Skip downloading and unpacking if the cache exists
        meta_yaml = outdir / "meta.yaml"
        if meta_yaml.exists():
            info = get_dict_from_cache(meta_yaml)
            if info is not None:
                return info

        # Download the file to an unique path
        filename = self.download(url, quiet=quiet)

        # Extract files from archived file
        return unpack(filename, outdir)


def str2bool(v) -> bool:
    return bool(strtobool(v))


def cmd_download(cmd=None):
    # espnet_model_zoo_download

    parser = argparse.ArgumentParser("Download file from Zenodo")
    parser.add_argument(
        "name",
        help="URL or model name in the form of <username>/<model name>. "
        "e.g. kamo-naoyuki/mini_an4_asr_train_raw_bpe_valid.acc.best",
    )
    parser.add_argument(
        "--cachedir",
        help="Specify cache dir. By default, download to module root.",
    )
    parser.add_argument(
        "--unpack",
        type=str2bool,
        default=False,
        help="Unpack the archived file after downloading.",
    )
    args = parser.parse_args(cmd)

    d = ModelDownloader(args.cachedir)
    if args.unpack:
        print(d.download_and_unpack(args.name))
    else:
        print(d.download(args.name))


def cmd_query(cmd=None):
    # espnet_model_zoo_query

    parser = argparse.ArgumentParser("Download file from Zenodo")

    parser.add_argument(
        "condition",
        nargs="*",
        default=[],
        help="Given desired condition in form of <key>=<value>. "
        "e.g. fs=16000. "
        "If no condition is given, you can view all available models",
    )
    parser.add_argument(
        "--key",
        default="name",
        help="The key name you want",
    )
    parser.add_argument(
        "--cachedir",
        help="Specify cache dir. By default, download to module root.",
    )
    args = parser.parse_args(cmd)

    conditions = dict(s.split("=") for s in args.condition)
    d = ModelDownloader(args.cachedir)
    for v in d.query(args.key, **conditions):
        print(v)



"""Upload files to Zenodo.

You need to do as follows in order to access zenodo:

1. Sign up to Zenodo: https://zenodo.org/
2. Create access_token: https://zenodo.org/account/settings/applications/tokens/new/
"""

import argparse
from datetime import datetime
from getpass import getpass
import json
import os
from pathlib import Path
import requests
from typing import Collection
from typing import Union

from espnet2.utils import config_argparse
from espnet2.utils.types import str2bool


class Zenodo:
    """Helper class to invoke Zenodo API

    REST API of zenodo: https://developers.zenodo.org/

    """

    def __init__(self, access_token: str, use_sandbox: bool = False):
        if use_sandbox:
            self.zenodo_url = "https://sandbox.zenodo.org"
        else:
            self.zenodo_url = "https://zenodo.org"

        self.params = {"access_token": access_token}
        self.headers = {"Content-Type": "application/json"}

    def create_deposition(self) -> requests.models.Response:
        r = requests.post(
            f"{self.zenodo_url}/api/deposit/depositions",
            params=self.params,
            json={},
            headers=self.headers,
        )
        if r.status_code != 201:
            raise RuntimeError(r.json()["message"])
        return r

    def get_deposition(self, r: Union[requests.models.Response, int]):
        if isinstance(r, requests.models.Response):
            deposition_id = r.json()["id"]
        else:
            deposition_id = r
        r = requests.get(
            f"{self.zenodo_url}/api/deposit/depositions/{deposition_id}",
            params=self.params,
            json={},
            headers=self.headers,
        )
        if r.status_code != 200:
            raise RuntimeError(r.json()["message"])
        return r

    def update_metadata(
        self, r: Union[requests.models.Response, int], data
    ) -> requests.models.Response:
        if isinstance(r, requests.models.Response):
            deposition_id = r.json()["id"]
        else:
            deposition_id = r

        r = requests.put(
            f"{self.zenodo_url}/api/deposit/depositions/{deposition_id}",
            params=self.params,
            data=json.dumps(data),
            headers=self.headers,
        )
        if r.status_code != 200:
            raise RuntimeError(r.json()["message"])
        return r

    def upload_file(
        self, r: Union[requests.models.Response, int], filename: Union[Path, str]
    ) -> requests.models.Response:
        if isinstance(r, int):
            r = requests.get(
                f"{self.zenodo_url}/api/deposit/depositions/{r}", headers=self.headers
            )

        bucket_url = r.json()["links"]["bucket"]
        name = Path(filename).name
        with open(filename, "rb") as fp:
            r = requests.put(
                f"{bucket_url}/{name}",
                data=fp,
                # No headers included since it's a raw byte request
                params=self.params,
            )
            if r.status_code != 200:
                raise RuntimeError(r.json()["message"])
        return r

    def publish(
        self, r: Union[requests.models.Response, int]
    ) -> requests.models.Response:
        if isinstance(r, requests.models.Response):
            deposition_id = r.json()["id"]
        else:
            deposition_id = r

        r = requests.post(
            f"{self.zenodo_url}/api/deposit/depositions/"
            f"{deposition_id}/actions/publish",
            params=self.params,
        )
        if r.status_code != 202:
            raise RuntimeError(r.json()["message"])
        return r


def upload(
    access_token: str,
    title: str,
    creator_name: str,
    description: str = "",
    files: Collection[Union[Path, str]] = (),
    affiliation: str = None,
    orcid: str = None,
    gnd: str = None,
    upload_type: str = "other",
    license: str = "CC-BY-4.0",
    keywords: Collection[str] = (),
    related_identifiers: Collection[dict] = (),
    community_identifer: str = None,
    use_sandbox: bool = True,
    publish: bool = False,
):
    zenodo = Zenodo(access_token, use_sandbox=use_sandbox)
    r = zenodo.create_deposition()

    # Update metatdata using old API
    creator = {"name": creator_name}
    if affiliation is not None:
        creator["affiliation"] = affiliation
    if orcid is not None:
        creator["orcid"] = orcid
    if gnd is not None:
        creator["gnd"] = gnd
    data = {
        "metadata": {
            "upload_type": upload_type,
            "publication_date": datetime.now().strftime("%Y-%m-%d"),
            "title": title,
            "description": description,
            "creators": [creator],
            "license": license,
            "keywords": list(keywords),
            "related_identifiers": list(related_identifiers),
        }
    }
    if community_identifer is not None:
        data["communities"] = [{"identifier": community_identifer}]
    zenodo.update_metadata(r, data)

    # Upload files using new API
    for f in files:
        # Check file existing
        if not Path(f).exists():
            raise FileNotFoundError(f"{f} is not found")
    for f in files:
        print(f"Now uploading {f}...")
        zenodo.upload_file(r, f)

    if publish:
        r = zenodo.publish(r)
        url = r.json()["links"]["latest_html"]
        print(f"Successfully published. Go to {url}")
    else:
        url = r.json()["links"]["html"]
        print(f"Successfully uploaded, but not published yet. Go to {url}")


def upload_espnet_model(
    access_token: str,
    title: str,
    creator_name: str,
    file: Collection[Union[Path, str]] = (),
    description: str = "",
    description_file: str = None,
    affiliation: str = None,
    license: str = "CC-BY-4.0",
    orcid: str = None,
    gnd: str = None,
    use_sandbox: bool = False,
    publish: bool = False,
):
    if description_file is not None:
        with open(description_file, "r", encoding="utf-8") as f:
            description = f.read()

    upload(
        access_token=access_token,
        title=title,
        description=description,
        creator_name=creator_name,
        files=file,
        keywords=[
            "ESPnet",
            "deep-learning",
            "python",
            "pytorch",
            "speech-recognition",
            "speech-synthesis",
            "speech-translation",
            "machine-translation",
        ],
        related_identifiers=[
            {
                "relation": "isSupplementTo",
                "identifier": "https://github.com/espnet/espnet",
            }
        ],
        affiliation=affiliation,
        license=license,
        orcid=orcid,
        gnd=gnd,
        use_sandbox=use_sandbox,
        publish=publish,
    )


def get_parser():
    parser = config_argparse.ArgumentParser(
        description="Upload files to Zenodo",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--access_token",
        help="Get your access_token from "
        "https://zenodo.org/account/settings/applications/ or  "
        "https://sandbox.zenodo.org/account/settings/applications/ . "
        "You can also give it from an environment variable 'ACCESS_TOKEN'",
    )
    parser.add_argument(
        "--title",
        required=True,
        help="e.g. ESPnet pretrained model, MT, "
        "Fisher-CallHome Spanish (Es->En), Transformer",
    )
    parser.add_argument("--creator_name", required=True, help="Your name")
    parser.add_argument("--file", nargs="+", required=True)
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--description", help="Give the description")
    group.add_argument("--description_file", help="Give the description from file")
    parser.add_argument(
        "--use_sandbox",
        type=str2bool,
        default=False,
        help="Use zenodo sandbox for testing",
    )
    parser.add_argument(
        "--publish", type=str2bool, default=False, help="Publish after uploading"
    )
    parser.add_argument("--license", default="CC-BY-4.0")
    parser.add_argument("--affiliation")
    parser.add_argument("--orcid")
    parser.add_argument("--gnd")
    return parser


def main(cmd=None):
    # espnet_model_zoo_upload

    parser = get_parser()
    args = parser.parse_args(cmd)

    # If --access_token is not given, get from "ACCESS_TOKEN"
    if args.access_token is None:
        args.access_token = os.environ.get("ACCESS_TOKEN")

    # If neither is given, input from stdin
    if args.access_token is None:
        if args.use_sandbox:
            zenodo_url = "https://sandbox.zenodo.org"
        else:
            zenodo_url = "https://zenodo.org"
        args.access_token = getpass(
            "Input Zenodo API Token\n"
            "(You can create it from "
            f"{zenodo_url}/account/settings/applications/tokens/new/): "
        )

    kwargs = vars(args)
    kwargs.pop("config")
    upload_espnet_model(**kwargs)


if __name__ == "__main__":
    main()
